# Copyright (C) 2022 Dremio Corporation

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import agate

from dbt.adapters.dremio.api.rest.endpoints import (
    sql_endpoint,
    job_status,
    job_results,
    job_cancel_api,
)
from dbt.adapters.dremio.api.parameters import Parameters

from dbt.events import AdapterLogger

logger = AdapterLogger("dremio")


class DremioCursor:
    def __init__(self, api_parameters: Parameters):
        self._parameters = api_parameters

        self._closed = False
        self._job_id = None
        self._rowcount = -1
        self._job_results = None
        self._table_results: agate.Table = None

    @property
    def parameters(self):
        return self._parameters

    @property
    def closed(self):
        return self._closed

    @closed.setter
    def closed(self, new_closed_value):
        self._closed = new_closed_value

    @property
    def rowcount(self):
        return self._rowcount

    @property
    def table(self) -> agate.Table:
        return self._table_results

    def job_results(self):
        if self.closed:
            raise Exception("CursorClosed")
        if self._job_results == None:
            self._populate_job_results()

        return self._job_results

    def job_cancel(self):
        # cancels current job
        logger.debug(f"Cancelling job {self._job_id}")
        return job_cancel_api(self._parameters, self._job_id)

    def close(self):
        if self.closed:
            raise Exception("CursorClosed")
        self._initialize()
        self.closed = True

    def execute(self, sql, bindings=None):
        if self.closed:
            raise Exception("CursorClosed")
        if bindings is None:
            self._initialize()

            json_payload = sql_endpoint(
                self._parameters, sql, context=None, ssl_verify=True
            )

            self._job_id = json_payload["id"]

            self._populate_rowcount()
            self._populate_job_results()
            self._populate_results_table()

        else:
            raise Exception("Bindings not currently supported.")

    def fetchone(self):
        row = None
        if self._table_results != None:
            row = self._table_results.rows[0]
        return row

    def fetchall(self):
        logger.debug(f"The fetch result is: {self._table_results.rows}")
        return self._table_results.rows

    def _initialize(self):
        self._job_id = None
        self._rowcount = -1
        self._table_results = None
        self._job_results = None

    def _populate_rowcount(self):
        if self.closed:
            raise Exception("CursorClosed")
        ## keep checking job status until status is one of COMPLETE, CANCELLED or FAILED
        ## map job results to AdapterResponse
        job_id = self._job_id

        last_job_state = ""
        job_status_response = job_status(self._parameters, job_id, ssl_verify=True)
        job_status_state = job_status_response["jobState"]

        while True:
            if job_status_state != last_job_state:
                logger.debug(f"Job State = {job_status_state}")
            if (
                job_status_state == "COMPLETED"
                or job_status_state == "CANCELLED"
                or job_status_state == "FAILED"
            ):
                break
            last_job_state = job_status_state
            job_status_response = job_status(self._parameters, job_id, ssl_verify=True)
            job_status_state = job_status_response["jobState"]

        # this is done as job status does not return a rowCount if there are no rows affected (even in completed job_state)
        # pyodbc Cursor documentation states "[rowCount] is -1 if no SQL has been executed or if the number of rows is unknown.
        # Note that it is not uncommon for databases to report -1 immediately after a SQL select statement for performance reasons."
        if "rowCount" not in job_status_response:
            rows = -1
            logger.debug("rowCount does not exist in job_status payload")
        else:
            rows = job_status_response["rowCount"]

        self._rowcount = rows

    def _populate_job_results(self):
        if self._job_results == None:
            self._job_results = job_results(
                self._parameters, self._job_id, offset=0, limit=100, ssl_verify=True
            )

    def _populate_results_table(self):
        if self._job_results != None:
            tester = agate.TypeTester()
            json_rows = self._job_results["rows"]
            self._table_results = json_rows
            for col in self._job_results["schema"]:
                name = col["name"]
                data_type_str = col["type"]["name"]
                if data_type_str == "BIGINT":
                    tester = agate.TypeTester(force={f"{name}": agate.Number()})

            self._table_results = agate.Table.from_object(
                json_rows, column_types=tester
            )



# Copyright (C) 2022 Dremio Corporation

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from dbt.adapters.base.relation import BaseRelation, Policy, ComponentName, RelationType
from typing import Optional, Tuple, Iterator


@dataclass
class DremioQuotePolicy(Policy):
    database: bool = True
    schema: bool = True
    identifier: bool = True


@dataclass
class DremioIncludePolicy(Policy):
    database: bool = True
    schema: bool = True
    identifier: bool = True


@dataclass(frozen=True, eq=False, repr=False)
class DremioRelation(BaseRelation):
    quote_policy: DremioQuotePolicy = DremioQuotePolicy()
    include_policy: DremioIncludePolicy = DremioIncludePolicy()
    no_schema = "no_schema"
    format: Optional[str] = None
    format_clause: Optional[str] = None

    def quoted_by_component(self, identifier, componentName):
        if componentName == ComponentName.Schema:
            return ".".join(self.quoted(folder) for folder in identifier.split("."))
        else:
            return self.quoted(identifier)

    def render(self) -> str:
        rendered = super().render()
        if self.format is not None and self.format_clause is not None:
            rendered = "".join(("table( ", rendered, " ( ", self.format_clause, " ) )"))
        return rendered

    def __post_init__(self):
        if self.path.schema is None:
            self.path.schema = DremioRelation.no_schema

    def _render_iterator(
        self,
    ) -> Iterator[Tuple[Optional[ComponentName], Optional[str]]]:

        for key in ComponentName:
            path_part: Optional[str] = None
            if self.include_policy.get_part(key):
                tmp_path_part = self.path.get_part(key)
                if not (
                    key == ComponentName.Schema and tmp_path_part == self.no_schema
                ):
                    path_part = tmp_path_part
                if path_part is not None and (
                    self.quote_policy.get_part(key)
                ):  # or key == ComponentName.Schema):
                    path_part = self.quoted_by_component(path_part, key)
            yield key, path_part



# Copyright (C) 2022 Dremio Corporation

# Copyright (c) 2019 Ryan Murray.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class DremioException(Exception):
    def __init__(self, msg, original_exception, response=None):
        message = f"{msg}: ({original_exception})"
        if response != None:
            message += f": ({response.text})"

        self.message = message
        self.original_exception = original_exception
        self.response = response

        super(DremioException, self).__init__(self.message)


class DremioUnauthorizedException(DremioException):
    pass


class DremioPermissionException(DremioException):
    pass


class DremioNotFoundException(DremioException):
    pass


class DremioBadRequestException(DremioException):
    pass


class DremioAlreadyExistsException(DremioException):
    pass


class DremioRequestTimeoutException(DremioException):
    pass


class DremioTooManyRequestsException(DremioException):
    pass


class DremioInternalServerException(DremioException):
    pass


class DremioServiceUnavailableException(DremioException):
    pass


class DremioGatewayTimeoutException(DremioException):
    pass



# Copyright (C) 2022 Dremio Corporation

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dbt.adapters.dremio.api.parameters import (
    Parameters,
    CloudParameters,
    SoftwareParameters,
)
from urllib.parse import quote


class UrlBuilder:
    SOFTWARE_LOGIN_ENDPOINT = "/apiv2/login"

    CLOUD_PROJECT_ENDPOINT = "/v0/projects"

    SOFTWARE_SQL_ENDPOINT = "/api/v3/sql"
    CLOUD_SQL_ENDPOINT = CLOUD_PROJECT_ENDPOINT + "/{}/sql"

    SOFTWARE_JOB_ENDPOINT = "/api/v3/job"
    CLOUD_JOB_ENDPOINT = CLOUD_PROJECT_ENDPOINT + "/{}/job"

    SOFTWARE_CATALOG_ENDPOINT = "/api/v3/catalog"
    CLOUD_CATALOG_ENDPOINT = CLOUD_PROJECT_ENDPOINT + "/{}/catalog"

    # https://docs.dremio.com/software/rest-api/jobs/get-job/
    OFFSET_DEFAULT = 0
    LIMIT_DEFAULT = 100

    # login_url only takes SoftwareParameters because Cloud uses pat.
    # There is no need to login to retrieve a token when Cloud only uses pat.
    @classmethod
    def login_url(cls, software_parameters: SoftwareParameters):
        return software_parameters.base_url + UrlBuilder.SOFTWARE_LOGIN_ENDPOINT

    @classmethod
    def sql_url(cls, parameters: Parameters):
        if type(parameters) is CloudParameters:
            return parameters.base_url + UrlBuilder.CLOUD_SQL_ENDPOINT.format(
                parameters.cloud_project_id
            )
        return parameters.base_url + UrlBuilder.SOFTWARE_SQL_ENDPOINT

    @classmethod
    def job_status_url(cls, parameters: Parameters, job_id):
        if type(parameters) is CloudParameters:
            return (
                parameters.base_url
                + UrlBuilder.CLOUD_JOB_ENDPOINT.format(parameters.cloud_project_id)
                + "/"
                + job_id
            )
        return parameters.base_url + UrlBuilder.SOFTWARE_JOB_ENDPOINT + "/" + job_id

    @classmethod
    def job_cancel_url(cls, parameters: Parameters, job_id):
        url_path = parameters.base_url
        if type(parameters) is CloudParameters:
            url_path += UrlBuilder.CLOUD_JOB_ENDPOINT
        else:
            url_path += UrlBuilder.SOFTWARE_JOB_ENDPOINT

        return url_path + "/{}/cancel".format(job_id)

    @classmethod
    def job_results_url(
        cls,
        parameters: Parameters,
        job_id,
        offset=OFFSET_DEFAULT,
        limit=LIMIT_DEFAULT,
    ):
        url_path = parameters.base_url
        if type(parameters) is CloudParameters:
            url_path += UrlBuilder.CLOUD_JOB_ENDPOINT.format(
                parameters.cloud_project_id
            )
        else:
            url_path += UrlBuilder.SOFTWARE_JOB_ENDPOINT

        return url_path + "/{}/results?offset={}&limit={}".format(job_id, offset, limit)

    @classmethod
    def catalog_url(cls, parameters: Parameters):
        url_path = parameters.base_url
        if type(parameters) is CloudParameters:
            url_path += UrlBuilder.CLOUD_CATALOG_ENDPOINT.format(
                parameters.cloud_project_id
            )
        else:
            url_path += UrlBuilder.SOFTWARE_CATALOG_ENDPOINT

        return url_path

    @classmethod
    def delete_catalog_url(cls, parameters: Parameters, cid):
        url_path = cls.catalog_url(parameters=parameters)
        url_path += f"/{cid}"

        return url_path

    @classmethod
    def catalog_item_by_id_url(cls, parameters: Parameters, catalog_id):
        url_path = parameters.base_url
        if type(parameters) is CloudParameters:
            url_path += UrlBuilder.CLOUD_CATALOG_ENDPOINT.format(
                parameters.cloud_project_id
            )
        else:
            url_path += UrlBuilder.SOFTWARE_CATALOG_ENDPOINT

        endpoint = "/{}".format(catalog_id)
        return url_path + endpoint

    @classmethod
    def catalog_item_by_path_url(cls, parameters: Parameters, path_list):
        url_path = parameters.base_url
        if type(parameters) is CloudParameters:
            url_path += UrlBuilder.CLOUD_CATALOG_ENDPOINT.format(
                parameters.cloud_project_id
            )
        else:
            url_path += UrlBuilder.SOFTWARE_CATALOG_ENDPOINT

        # Escapes special characters
        quoted_path_list = [quote(i, safe="") for i in path_list]
        # Converts list to string separated by '/'
        joined_path_str = "/".join(quoted_path_list).replace('"', "")
        endpoint = f"/by-path/{joined_path_str}"
        return url_path + endpoint



# Copyright (C) 2022 Dremio Corporation

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from dataclasses import dataclass
from typing import Optional
from dbt.adapters.dremio.api.authentication import DremioAuthentication
from dbt.adapters.dremio.credentials import DremioCredentials


@dataclass
class Parameters:
    base_url: str
    authentication: DremioAuthentication


@dataclass
class CloudParameters(Parameters):
    cloud_project_id: str


@dataclass
class SoftwareParameters(Parameters):
    pass


@dataclass
class ParametersBuilder:
    authentication: Optional[DremioCredentials] = None

    @classmethod
    def build(cls, credentials: DremioCredentials):
        if (
            credentials.cloud_host != None
            and credentials.software_host == None
            and credentials.cloud_project_id != None
        ):
            return CloudParametersBuilder(
                cls._build_dremio_authentication(credentials=credentials),
                credentials.cloud_host,
                credentials.cloud_project_id,
            )
        elif (
            credentials.software_host != None
            and credentials.cloud_host == None
            and credentials.port != None
            and credentials.use_ssl != None
        ):
            return SoftwareParametersBuilder(
                cls._build_dremio_authentication(credentials=credentials),
                credentials.software_host,
                credentials.port,
                credentials.use_ssl,
            )
        raise ValueError("Credentials match neither Cloud nor Software")

    @abstractmethod
    def build_base_url(self):
        pass

    @abstractmethod
    def get_parameters(self) -> Parameters:
        pass

    @classmethod
    def _build_dremio_authentication(self, credentials: DremioCredentials):
        return DremioAuthentication.build(
            credentials.UID, credentials.PWD, credentials.pat
        )


@dataclass
class CloudParametersBuilder(ParametersBuilder):
    cloud_host: str = None
    cloud_project_id: str = None

    def build_base_url(self):
        protocol = "https"
        base_url = f"{protocol}://{self.cloud_host}"
        return base_url

    def get_parameters(self) -> Parameters:
        return CloudParameters(
            base_url=self.build_base_url(),
            authentication=self.authentication,
            cloud_project_id=self.cloud_project_id,
        )


@dataclass
class SoftwareParametersBuilder(ParametersBuilder):
    software_host: str = None
    port: str = None
    use_ssl: bool = None

    def build_base_url(self):
        protocol = "http"
        if self.use_ssl:
            protocol = "https"
        base_url = f"{protocol}://{self.software_host}:{self.port}"
        return base_url

    def get_parameters(self) -> Parameters:
        return SoftwareParameters(
            base_url=self.build_base_url(), authentication=self.authentication
        )



# Copyright (C) 2022 Dremio Corporation

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import agate
from typing import Tuple, Optional
from contextlib import contextmanager

from typing import List
from dbt.adapters.dremio.api.cursor import DremioCursor
from dbt.adapters.dremio.api.handle import DremioHandle
from dbt.adapters.dremio.api.parameters import ParametersBuilder

import time
import json

import dbt.exceptions
from dbt.adapters.sql import SQLConnectionManager
from dbt.contracts.connection import AdapterResponse

from dbt.adapters.dremio.credentials import DremioCredentials

from dbt.adapters.dremio.api.rest.endpoints import (
    delete_catalog,
    create_catalog_api,
    get_catalog_item,
)
from dbt.adapters.dremio.api.rest.error import (
    DremioAlreadyExistsException,
    DremioNotFoundException,
    DremioRequestTimeoutException,
    DremioTooManyRequestsException,
    DremioInternalServerException,
    DremioServiceUnavailableException,
    DremioGatewayTimeoutException,
)

from dbt.events import AdapterLogger

logger = AdapterLogger("dremio")


class DremioConnectionManager(SQLConnectionManager):
    TYPE = "dremio"
    DEFAULT_CONNECTION_RETRIES = 1

    retries = DEFAULT_CONNECTION_RETRIES

    @contextmanager
    def exception_handler(self, sql):
        try:
            yield
        except Exception as e:
            logger.debug(f"Error running SQL: {sql}")
            self.release()
            if isinstance(e, dbt.exceptions.RuntimeException):
                # during a sql query, an internal to dbt exception was raised.
                # this sounds a lot like a signal handler and probably has
                # useful information, so raise it without modification.
                raise

            raise dbt.exceptions.RuntimeException(e)

    @classmethod
    def open(cls, connection):

        if connection.state == "open":
            logger.debug("Connection is already open, skipping open.")
            return connection

        credentials = connection.credentials
        parameters_builder = ParametersBuilder.build(credentials)
        api_parameters = parameters_builder.get_parameters()

        def connect():
            handle = DremioHandle(api_parameters)
            _ = handle.cursor()
            connection.state = "open"
            connection.handle = handle
            logger.debug(f"Connected to db: {credentials.database}")
            return handle

        retryable_exceptions = [
            # list of retryable_exceptions underlying driver might expose
            DremioRequestTimeoutException,
            DremioTooManyRequestsException,
            DremioInternalServerException,
            DremioServiceUnavailableException,
            DremioGatewayTimeoutException,
        ]

        def exponential_backoff_retry_timeout(retries: int) -> int:
            BASE = 2  # multiplicative factor
            time_delay = pow(BASE, retries)
            return time_delay

        return cls.retry_connection(
            connection,
            connect=connect,
            logger=logger,
            retry_limit=cls.retries,
            retry_timeout=exponential_backoff_retry_timeout,
            retryable_exceptions=retryable_exceptions,
        )

    @classmethod
    def is_cancelable(cls) -> bool:
        return True

    def cancel(self, connection):
        return connection.handle.cursor.job_cancel()

    def commit(self, *args, **kwargs):
        pass

    def rollback(self, *args, **kwargs):
        pass

    def add_begin_query(self):
        pass

    def add_commit_query(self):
        pass

    # Auto_begin may not be relevant with the rest_api
    def add_query(self, sql, auto_begin=True, bindings=None, abridge_sql_log=False):

        connection = self.get_thread_connection()
        if auto_begin and connection.transaction_open is False:
            self.begin()

        logger.debug(f'Using {self.TYPE} connection "{connection.name}')

        with self.exception_handler(sql):
            if abridge_sql_log:
                logger.debug("On {}: {}....".format(connection.name, sql[0:512]))
            else:
                logger.debug("On {}: {}".format(connection.name, sql))

            pre = time.time()
            cursor = connection.handle.cursor()

            if bindings is None:
                cursor.execute(sql)
            else:
                logger.debug(f"Bindings: {bindings}")
                cursor.execute(sql, bindings)

            logger.debug(
                "SQL status: {} in {:0.2f} seconds".format(
                    self.get_response(cursor), (time.time() - pre)
                )
            )
            return connection, cursor

    @classmethod
    def get_credentials(cls, credentials):
        return credentials

    @classmethod
    def get_response(cls, cursor: DremioCursor) -> AdapterResponse:
        rows = cursor.rowcount
        message = "OK" if rows == -1 else str(rows)
        return AdapterResponse(_message=message, rows_affected=rows)

    def execute(
        self, sql: str, auto_begin: bool = False, fetch: bool = False
    ) -> Tuple[AdapterResponse, agate.Table]:
        sql = self._add_query_comment(sql)
        _, cursor = self.add_query(sql, auto_begin)
        response = self.get_response(cursor)
        fetch = True
        if fetch:
            table = cursor.table
        else:
            table = dbt.clients.agate_helper.empty_table()

        return response, table

    def drop_catalog(self, database, schema):
        logger.debug('Dropping schema "{}.{}"', database, schema)

        thread_connection = self.get_thread_connection()
        connection = self.open(thread_connection)
        credentials = connection.credentials
        api_parameters = connection.handle.get_parameters()

        path_list = self._create_path_list(database, schema)
        if database != credentials.datalake:
            try:
                catalog_info = get_catalog_item(
                    api_parameters,
                    catalog_id=None,
                    catalog_path=path_list,
                    ssl_verify=False,
                )
            except DremioNotFoundException:
                logger.debug("Catalog not found. Returning")
                return

            delete_catalog(api_parameters, catalog_info["id"], ssl_verify=False)

    def create_catalog(self, database, schema):
        thread_connection = self.get_thread_connection()
        connection = self.open(thread_connection)
        credentials = connection.credentials
        api_parameters = connection.handle.get_parameters()

        if database == ("@" + credentials.UID):
            logger.debug("Database is default: creating folders only")
        else:
            self._create_space(database, api_parameters)
        if database != credentials.datalake:
            self._create_folders(database, schema, api_parameters)
        return

    def _make_new_space_json(self, name) -> json:
        python_dict = {"entityType": "space", "name": name}
        return json.dumps(python_dict)

    def _make_new_folder_json(self, path) -> json:
        python_dict = {"entityType": "folder", "path": path}
        return json.dumps(python_dict)

    def _create_space(self, database, api_parameters):
        space_json = self._make_new_space_json(database)
        try:
            create_catalog_api(api_parameters, space_json, False)
        except DremioAlreadyExistsException:
            logger.debug(f"Database {database} already exists. Creating folders only.")

    def _create_folders(self, database, schema, api_parameters):
        temp_path_list = [database]
        for folder in schema.split("."):
            temp_path_list.append(folder)
            folder_json = self._make_new_folder_json(temp_path_list)
            try:
                create_catalog_api(api_parameters, folder_json, False)
            except DremioAlreadyExistsException:
                logger.debug(f"Folder {folder} already exists.")

    def _create_path_list(self, database, schema):
        path = [database]
        folders = schema.split(".")
        path.extend(folders)
        return path



# Copyright (C) 2022 Dremio Corporation

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import agate
from dbt.adapters.sql import SQLAdapter
from dbt.adapters.dremio import DremioConnectionManager
from dbt.adapters.dremio.relation import DremioRelation
from typing import Dict

from typing import List
from typing import Optional
from dbt.adapters.base.relation import BaseRelation
from dbt.adapters.sql.impl import DROP_RELATION_MACRO_NAME
from dbt.events import AdapterLogger

logger = AdapterLogger("dremio")


class DremioAdapter(SQLAdapter):
    ConnectionManager = DremioConnectionManager
    Relation = DremioRelation

    @classmethod
    def date_function(cls):
        return "current_date"

    @classmethod
    def convert_text_type(cls, agate_table, col_idx):
        return "varchar"

    @classmethod
    def convert_datetime_type(cls, agate_table, col_idx):
        return "timestamp"

    @classmethod
    def convert_date_type(cls, agate_table, col_idx):
        return "date"

    @classmethod
    def convert_boolean_type(cls, agate_table, col_idx):
        return "boolean"

    @classmethod
    def convert_number_type(cls, agate_table, col_idx):
        decimals = agate_table.aggregate(agate.MaxPrecision(col_idx))
        return "decimal" if decimals else "bigint"

    @classmethod
    def convert_time_type(cls, agate_table, col_idx):
        return "time"

    def create_schema(self, relation: DremioRelation) -> None:
        database = relation.database
        schema = relation.schema
        self.connections.create_catalog(database, schema)

    def drop_schema(self, relation: DremioRelation) -> None:
        if relation.type == "table":
            self.execute_macro(DROP_RELATION_MACRO_NAME, kwargs={"relation": relation})

        else:
            database = relation.database
            schema = relation.schema
            self.connections.drop_catalog(database, schema)

    def timestamp_add_sql(
        self, add_to: str, number: int = 1, interval: str = "hour"
    ) -> str:
        return f"DATE_ADD({add_to}, CAST({number} AS INTERVAL {interval}))"

    def get_rows_different_sql(
        self,
        relation_a: BaseRelation,
        relation_b: BaseRelation,
        column_names: Optional[List[str]] = ["*"],
        except_operator: str = "EXCEPT",
    ) -> str:
        """Generate SQL for a query that returns a single row with a two
        columns: the number of rows that are different between the two
        relations and the number of mismatched rows.
        """
        # This method only really exists for test reasons.
        names: List[str]
        if column_names is None:
            columns = self.get_columns_in_relation(relation_a)
            names = sorted((self.quote(c.name) for c in columns))
        else:
            names = sorted((self.quote(n) for n in column_names))
        columns_csv = ", ".join(names)

        sql = COLUMNS_EQUAL_SQL.format(
            columns=columns_csv,
            relation_a=str(relation_a),
            relation_b=str(relation_b),
            except_op=except_operator,
        )

        return sql

    def standardize_grants_dict(self, grants_table: agate.Table) -> dict:
        """Translate the result of `show grants` (or equivalent) to match the
        grants which a user would configure in their project.

        Ideally, the SQL to show grants should also be filtering:
        filter OUT any grants TO the current user/role (e.g. OWNERSHIP).
        If that's not possible in SQL, it can be done in this method instead.

        :param grants_table: An agate table containing the query result of
            the SQL returned by get_show_grant_sql
        :return: A standardized dictionary matching the `grants` config
        :rtype: dict
        """
        grants_dict: Dict[str, List[str]] = {}
        for row in grants_table:
            # Just needed to change these two values to match Dremio cols
            grantee = row["grantee_id"]
            privilege = row["privilege"]
            if privilege in grants_dict.keys():
                grants_dict[privilege].append(grantee)
            else:
                grants_dict.update({privilege: [grantee]})
        return grants_dict


COLUMNS_EQUAL_SQL = """
with diff_count as (
    SELECT
        1 as id,
        COUNT(*) as num_missing FROM (
            (SELECT {columns} FROM {relation_a} {except_op}
             SELECT {columns} FROM {relation_b})
             UNION ALL
            (SELECT {columns} FROM {relation_b} {except_op}
             SELECT {columns} FROM {relation_a})
        ) as a
), table_a as (
    SELECT COUNT(*) as num_rows FROM {relation_a}
), table_b as (
    SELECT COUNT(*) as num_rows FROM {relation_b}
), row_count_diff as (
    select
        1 as id,
        table_a.num_rows - table_b.num_rows as difference
    from table_a, table_b
)
select
    row_count_diff.difference as row_count_difference,
    diff_count.num_missing as num_mismatched
from row_count_diff
join diff_count using (id)
""".strip()



# Copyright (C) 2022 Dremio Corporation

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# __init__.py
from .rest.endpoints import (
    delete_catalog,
    sql_endpoint,
    job_status,
    create_catalog_api,
    get_catalog_item,
    login,
)
